# pylint: disable=R0903
# coding: utf-8
from copr.util import UnicodeMixin


class EntityTypes(object):
    ROOT = "root"
    PROJECT = "project"
    PROJECT_CHROOT = "project_chroot"
    BUILD = "build"
    BUILD_TASK = "build_task"
    MOCK_CHROOT = "mock_chroot"


class BuiltPackage(UnicodeMixin):

    def __init__(self, name, version):
        self.name = name
        self.version = version

    def __unicode__(self):
        return u"{0} {1}".format(self.name, self.version)


class BuildStateValues(object):
    FAILED = "failed"
    SUCCEEDED = "succeeded"
    CANCELED = "canceled"
    RUNNING = "running"
    PENDING = "pending"
    SKIPPED = "skipped"
    STARTING = "starting"
    IMPORTING = "importing"
    FORKED = "forked"
    WAITING = "waiting"
    UNKNOWN = "unknown"

ALLOWED_BUILD_STATES = set([
    BuildStateValues.FAILED,
    BuildStateValues.SUCCEEDED,
    BuildStateValues.CANCELED,
    BuildStateValues.RUNNING,
    BuildStateValues.PENDING,
    BuildStateValues.SKIPPED,
    BuildStateValues.STARTING,
    BuildStateValues.IMPORTING,
    BuildStateValues.FORKED,
    BuildStateValues.WAITING,
    BuildStateValues.UNKNOWN,
])



from munch import Munch


class CoprException(Exception):
    """
    Base Copr exception
    """
    def __init__(self, msg=None, response=None):
        super(CoprException, self).__init__(msg)
        self.result = Munch(error=msg, __response__=response)


class CoprRequestException(CoprException):
    """
    Raised when the API request doesn't proceed successfully
    """
    pass


class CoprNoResultException(CoprException):
    """
    Raised when no result data is returned
    """
    pass


class CoprTimeoutException(CoprException):
    """
    Raised when the API request timeouted
    """
    pass


class CoprValidationException(CoprException):
    """
    Raised when the data sent from client to API are not valid
    """
    pass


class CoprNoConfigException(CoprException):
    """
    Exception thrown when no config file is found
    """
    pass


class CoprConfigException(CoprException):
    """
    Exception thrown when the config file is incomplete or malformed.
    """
    pass



# coding: utf-8
import time
from copr.client_v2.resources import Build


def wait_for_builds(builds, check_timeout=30):
    """
    :type builds: list of Build
    :param time check_timeout: delay between status check attempts
    """

    if isinstance(builds, Build):
        builds = [builds]

    working = list(builds)
    while len(working) > 0:
        next_gen = []
        for b in working:
            if not b.get_self().is_finished():
                next_gen.append(b)
        working = next_gen
        time.sleep(check_timeout)



import os

from requests.models import Response

from copr import CoprClient

try:
     from unittest import mock
except ImportError:
     # Python 2 version depends on mock
     import mock

path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)
resource_location = os.path.join(dir_path, "resources")
config_location = os.path.join(resource_location, "copr_cli.conf")


def test_client_from_dict():
    cl = CoprClient(
        login="api-login",
        username="user_name",
        token="api-token",
        copr_url="http://copr-fe-dev.cloud.fedoraproject.org"
    )

    assert isinstance(cl, CoprClient)
    assert cl.login == "api-login"
    assert cl.token == "api-token"
    assert cl.username == "user_name"


def test_client_from_config():
    cl = CoprClient.create_from_file_config(config_location)
    assert isinstance(cl, CoprClient)
    assert cl.login == "api-login"
    assert cl.token == "api-token"
    assert cl.username == "user_name"


def test_list_projects():
    CoprClient.create_from_file_config(config_location)


# TODO: package https://github.com/dropbox/responses and use it
def make_mock_response(filename, status_code=None):
    response = Response()
    response.status_code = status_code or 200
    response.encoding = "utf-8"
    with open(os.path.join(resource_location, filename)) as text:
        response._content = text.read().encode()
    return response


@mock.patch('requests.request')
def test_projects_list(mock_request):
    mock_client = CoprClient.create_from_file_config(config_location)
    mock_request.return_value = make_mock_response("projects_list.200.json")

    test_resp = mock_client.get_projects_list()
    assert len(test_resp.projects_list) == 1
    test_project = test_resp.projects_list[0]
    assert test_project.projectname == "perl516-el7"
    assert test_project.description == "Test description"


@mock.patch('requests.request')
def test_get_build_status(mock_request):
    mock_client = CoprClient.create_from_file_config(config_location)

    mock_request.return_value = make_mock_response("build_details.200.json")

    test_resp = mock_client.get_build_details(27382)

    assert test_resp.status == "succeeded"
    assert test_resp.project == "atomic-next"
    assert test_resp.built_pkgs == [u'golang-github-stretchr-objx-devel 0']
    assert test_resp.submitted_on == 1408031345


@mock.patch('requests.request')
def test_get_build_config(mock_request):
    mock_client = CoprClient.create_from_file_config(config_location)
    mock_request.return_value = make_mock_response("build_config.200.json")
    test_resp = mock_client.get_build_config("project", "chroot")
    assert 'additional_packages' in test_resp.data
    expected_args = ["additional_packages", "additional_repos", "chroot",
                     "enable_net", "repos", "use_bootstrap_container",
                     "with_opts", "without_opts"]
    for arg in expected_args:
        assert arg in test_resp.data
        test_resp.data.pop(arg)
    assert not test_resp.data.keys()



# coding: utf-8
# pylint: disable=W1202
from collections import namedtuple
import json

from logging import getLogger

from requests import request, ConnectionError
from requests_toolbelt.multipart.encoder import MultipartEncoder

from ..util import UnicodeMixin

log = getLogger(__name__)


class RequestError(Exception, UnicodeMixin):
    def __init__(self, msg, url, request_kwargs=None, response=None, request_body=None):
        super(RequestError, self).__init__()
        self.msg = msg
        self.url = url
        self.request_body = request_body
        self.request_kwargs = request_kwargs or dict()
        if "auth" in self.request_kwargs:
            self.request_kwargs["auth"] = "<hidden>"
        self.response = response

    @property
    def response_json(self):
        if self.response is None:
            raise ValueError("No response")
        if self.response.headers["content-type"] == "application/json":
            try:
                result = json.loads(self.response.text)
            except (ValueError, AttributeError):
                raise ValueError(
                    "Malformed response, couldn't "
                    "get json content, raw:\n{0}"
                    .format(self.response.text)
                )

            return result
        else:
            return None

    def __unicode__(self):
        res = "Error occurred while accessing {0}: {1}\n".format(
            self.url, self.msg)
        if self.response is not None:
            res += "code {0}: {1}\n".format(self.response.status_code, self.response_json["message"])
        return res


class NetworkError(RequestError):
    def __init__(self, url, request_kwargs, requests_error):
        self.requests_error = requests_error
        super(NetworkError, self).__init__(
            u"Connection error", url, request_kwargs)

    def __unicode__(self):
        res = super(NetworkError, self).__unicode__()
        res += u"Original error: {0}\n".format(self.requests_error)
        return res


class AuthError(RequestError):
    def __init__(self, url, request_kwargs, response):
        super(AuthError, self).__init__("Authorization failed",
                                        url, request_kwargs, response)


class ResponseWrapper(object):

    def __init__(self, response):
        """
        :raises ValueError: when fails to deserialize json content
        """
        self.response = response
        if response.status_code != 204 and response.content:
            if isinstance(response.content, bytes):
                self.json = json.loads(response.content.decode('utf-8'))
            else:
                self.json = json.loads(response.content)
        else:
            self.json = None

    @property
    def status_code(self):
        return self.response.status_code

    @property
    def headers(self):
        return self.response.headers

MultiPartTuple = namedtuple("MultiPartTuple", ["key", "name", "obj", "content_type"])


class NetClient(object):
    """
    Abstraction around python-requests

    :param str login: login for BasicAuth
    :param str password: password for BasicAuth
    """

    def __init__(self, login=None, password=None):
        self.login = login
        self.token = password

    def request_multipart(self, url, method=None, query_params=None,
                          data_parts=None, do_auth=False):
        """
        :type data_parts: list of MultiPartTuple

        """
        parts = {}
        for key, name, obj, content_type in data_parts:
            parts[key] = (name, obj, content_type)

        data = MultipartEncoder(parts)
        headers = {
            "content-type": data.content_type
        }
        return self.request(url, method=method, query_params=query_params,
                            data=data, do_auth=do_auth, headers=headers)

    def request(self, url, method=None, query_params=None, data=None, do_auth=False, headers=None):
        """
        :param str method: what HTTP method to use, default is GET, allowed methods: GET, POST, PUT, DELETE
        :param dict query_params: HTTP query parameters
        :param str data: serialized data, when present set default content type to application/json
        :param dict headers: dict with headers, takes priority over implicit ones
        :param bool do_auth: sends auth headers when enabled

        :raises: RequestError
        """

        if method is None:
            method = "get"
        elif method.lower() not in ["get", "post", "delete", "put"]:
            raise RequestError("Method {0} not allowed".format(method), url)

        kwargs = {}
        headers = headers or {}
        if do_auth:
            if self.login is None or self.token is None:
                raise RequestError("Credentionals for BasicAuth "
                                   "not set, request aborted",
                                   url, kwargs)
            kwargs["auth"] = (self.login, self.token)
        if query_params:
            kwargs["params"] = query_params
        if data:
            kwargs["data"] = data
            if "content-type" not in headers:
                headers["content-type"] = "application/json"

        try:
            response = request(
                method=method.upper(),
                url=url,
                headers=headers,
                **kwargs
            )
            log.debug("raw response: {0}".format(response.text))
        except ConnectionError as e:
            raise NetworkError(url, kwargs, e)

        if response.status_code == 403:
            raise AuthError(url, kwargs, response)

        if response.status_code >= 500:
            raise RequestError("Server error", url, kwargs, response)

        if response.status_code > 399:
            raise RequestError("Request error", url, kwargs, response)

        try:
            return ResponseWrapper(response)
        except ValueError:
            raise RequestError("Failed to parse server response", url, kwargs, response)

    def get(self, url, query_params=None):
        return self.request(url, query_params=query_params)



# coding: utf-8
# pylint: disable=E1101, C0102
import pytest
from copr.client_v2.schemas import Schema, fields
from copr.client_v2.entities import Entity


class FooSchema(Schema):
    foo = fields.Str()
    bar = fields.Int()


class FooEntity(Entity):
    _schema = FooSchema()


class TestEntities(object):

    entity = None

    # pylint: disable=W0613
    def setup_method(self, method):
        self.entity = FooEntity.from_dict({"foo": "baz", "bar": 123})

    def test_from_dict(self):
        assert self.entity.foo == "baz"
        assert self.entity.bar == 123
        assert not hasattr(self.entity, "non_existing_attribute")

        with pytest.raises(AttributeError):
            assert self.entity.non_existing_attribute

    def test_to_dict(self):
        assert set(self.entity.to_dict().items()) == \
               set([("foo", "baz"), ("bar", 123)])

    def test_to_json(self):
        assert self.entity.to_json() == '{"foo": "baz", "bar": 123}' \
            or self.entity.to_json() == '{"bar": 123, "foo": "baz"}'




# pylint: disable=R0903
# coding: utf-8

import six

if six.PY2:
    from collections import Iterable
else:
    from collections.abc import Iterable

from six import with_metaclass

from ..util import UnicodeMixin

from .entities import Link, ProjectEntity, ProjectChrootEntity, BuildEntity, MockChrootEntity, BuildTaskEntity
from .schemas import EmptySchema, BuildSchema, ProjectSchema, ProjectChrootSchema, MockChrootSchema, BuildTaskSchema

try:
    _ = ProjectSchema(strict=True)
    kwargs = {"strict": True}
except TypeError:
    kwargs = {}


class EntityFieldDescriptor(object):
    """
    Entity Field Descriptor
    """
    def __init__(self, name):
        self.name = name

    def __get__(self, obj, objtype):
        """
        :type obj: IndividualResource
        """
        return getattr(obj._entity, self.name)

    def __set__(self, obj, val):
        """
        :type obj: IndividualResource
        """
        setattr(obj._entity, self.name, val)


class ReadOnlyFieldDescriptor(object):
    """
    Entity read-only Field Descriptor
    """
    def __init__(self, name):
        self.name = name

    def __get__(self, obj, objtype):
        """
        :type obj: IndividualResource
        """
        return getattr(obj._entity, self.name)


class EntityFieldsMetaClass(type):
    """
    Magic: we take fields info from class._schema and attach EntityFieldDescriptor
    to Resource classes
    """
    def __new__(mcs, class_name, bases, class_attrs):
        schema = class_attrs.get("_schema")
        if schema:
            for f_name, f in schema.fields.items():
                class_attrs[f_name] = EntityFieldDescriptor(f_name)

        entity_methods = []
        for base in bases:
            entity_methods.extend(getattr(base, "_entity_methods", []))
        entity_methods.extend(class_attrs.get("_entity_methods", []))

        for m_name in entity_methods:
            class_attrs[m_name] = ReadOnlyFieldDescriptor(m_name)

        return type.__new__(mcs, class_name, bases, class_attrs)


# pylint: disable=E1101
class IndividualResource(with_metaclass(EntityFieldsMetaClass, UnicodeMixin)):
    """
    :type handle: client_v2.handlers.AbstractHandle or None
    :type response: ResponseWrapper or None
    :type links: (dict of (str, Link)) or None
    """
    _schema = EmptySchema(**kwargs)

    # todo:  this methods only serialize fields which can be modified by the user
    # think about an approach to override `load_only=True` fields in
    #  our schemas during the dump function
    # _entity_methods = ["to_json", "to_dict"]

    def __init__(self, entity, handle=None, response=None, links=None, embedded=None, options=None):

        self._entity = entity
        self._handle = handle
        self._response = response
        self._links = links
        self._embedded = embedded or dict()
        self._options = options or dict()

    def __dir__(self):
        res = list(set(
            dir(self.__class__) + list(self.__dict__.keys())
        ))
        if self._entity:
            res.extend(self._schema.fields.keys())
        return res

    def __unicode__(self):
        return self._entity.__unicode__()

    def get_href_by_name(self, name):
        """
        :type name: str
        """
        return self._links[name].href


class Root(IndividualResource):

    def __init__(self, response, links, root_url):
        super(Root, self).__init__(entity=None, response=response, links=links)
        self.root_url = root_url

    def get_resource_base_url(self, resource_name):
        """
        :param str resource_name:
        """
        return "{0}{1}".format(self.root_url, self.get_href_by_name(resource_name))

    @classmethod
    def from_response(cls, response, root_url):
        """
        :type response: ResponseWrapper
        :type root_url: unicode
        """
        data_dict = response.json
        links = Link.from_dict(data_dict["_links"])
        return Root(response=response, links=links, root_url=root_url)


class Build(IndividualResource):
    """
    :type entity: BuildEntity
    :type handle: copr.client_v2.handlers.BuildHandle
    """
    _schema = BuildSchema(**kwargs)
    _entity_methods = ["is_finished"]

    def __init__(self, entity, handle, **kwargs):
        super(Build, self).__init__(entity=entity, handle=handle, **kwargs)
        self._entity = entity
        self._handle = handle

    @classmethod
    def from_response(cls, handle, data_dict, response=None, options=None):
        links = Link.from_dict(data_dict["_links"])
        entity = BuildEntity.from_dict(data_dict["build"])
        return cls(entity=entity, handle=handle,
                   response=response, links=links, options=options)

    def get_self(self):
        """ Retrieves fresh build object from the service

        :rtype: :py:class:`~.Build`
        """
        return self._handle.get_one(self.id)

    def cancel(self):
        """ Updates the current build

        :rtype: :py:class:`.OperationResult`
        """
        return self._handle.cancel(self._entity)

    def delete(self):
        """ Deletes the current build

        :rtype: :py:class:`.OperationResult`
        """
        return self._handle.delete(self.id)

    def get_build_tasks(self, **query_options):
        """ Get build tasks owned by this build

        :param query_options: see :py:meth:`.handlers.BuildHandle.get_list`
        :rtype: :py:class:`~.BuildTasksList`
        """
        handle = self._handle.get_build_tasks_handle()
        return handle.get_list(build_id=self.id, **query_options)


class BuildTask(IndividualResource):
    """
    :type entity: BuildTaskEntity
    :type handle: copr.client_v2.handlers.BuildTaskHandle
    """
    _schema = BuildTaskSchema(**kwargs)

    def __init__(self, entity, handle, **kwargs):
        super(BuildTask, self).__init__(entity=entity, handle=handle, **kwargs)
        self._entity = entity
        self._handle = handle

    def get_self(self):
        """ Retrieves fresh build task object from the service

        :rtype: :py:class:`~.Build`
        """
        return self._handle.get_one(self.build_id, self.chroot_name)


    @classmethod
    def from_response(cls, handle, data_dict, response=None, options=None):
        links = Link.from_dict(data_dict["_links"])
        entity = BuildTaskEntity.from_dict(data_dict["build_task"])
        return cls(entity=entity, handle=handle,
                   response=response, links=links, options=options)


class Project(IndividualResource):
    """
    :type entity: ProjectEntity
    :type handle: copr.client_v2.handlers.ProjectHandle
    """
    _schema = ProjectSchema(**kwargs)

    def __init__(self, entity, handle, **kwargs):
        super(Project, self).__init__(entity=entity, handle=handle, **kwargs)

        self._entity = entity
        self._handle = handle

    def update(self):
        """ Updates project using the current state.

        Shortcut for for :py:meth:`.ProjectHandle.update`

        :rtype: :py:class:`.OperationResult`
        """
        return self._handle.update(self._entity)

    def delete(self):
        """ Updates project using the current state

        :rtype: :py:class:`.OperationResult`
        """
        return self._handle.delete(self.id)

    def get_self(self):
        """ Retrieves fresh project object from the service

        :rtype: :py:class:`.Project`
        """
        return self._handle.get_one(self.id)

    def get_builds(self, **query_options):
        """ Get builds owned by this project

        :param query_options: see :py:meth:`.handlers.BuildHandle.get_list`
        :rtype: :py:class:`~.BuildsList`
        """
        handle = self._handle.get_builds_handle()
        return handle.get_list(project_id=self.id, **query_options)

    def get_build_tasks(self, **query_options):
        """ Get build tasks owned by this project

        :param query_options: see :py:meth:`.handlers.BuildHandle.get_list`
        :rtype: :py:class:`~.BuildTasksList`
        """
        handle = self._handle.get_build_tasks_handle()
        return handle.get_list(project_id=self.id, **query_options)

    def get_project_chroot(self, name):
        """ Retrieves project chroot object by the given name

        :param str name: mock chroot name
        :rtype: :py:class:`.ProjectChroot`
        """
        handle = self._handle.get_project_chroots_handle()
        return handle.get_one(self, name)

    def get_project_chroot_list(self):
        """ Retrieves project chroots list

        :rtype: :py:class:`.ProjectChrootList`
        """
        handle = self._handle.get_project_chroots_handle()
        return handle.get_list(self)

    def enable_project_chroot(self, name):
        """
        Enables given chroot for this project

        Shortcut for for :py:meth:`.ProjectChrootHandle.enable`

        :param str name: mock chroot name
        :rtype: :py:class:`.OperationResult`
        """
        handle = self._handle.get_project_chroots_handle()
        return handle.enable(self, name)

    def create_build_from_file(self, *args, **kwargs):
        """
        Shortcut for :py:meth:`.BuildHandle.create_from_file`
        (here you don't need to specify project_id)
        """
        builds = self._handle.get_builds_handle()
        return builds.create_from_file(self.id, *args, **kwargs)

    def create_build_from_url(self, *args, **kwargs):
        """
        Shortcut for :py:meth:`.BuildHandle.create_from_file`
        (here you don't need to specify project_id)
        """
        builds = self._handle.get_builds_handle()
        return builds.create_from_url(self.id, *args, **kwargs)

    @classmethod
    def from_response(cls, handle, data_dict, response=None, options=None):
        links = Link.from_dict(data_dict["_links"])
        entity = ProjectEntity.from_dict(data_dict["project"])
        return cls(entity=entity, handle=handle,
                   response=response, links=links, options=options)


class ProjectChroot(IndividualResource):
    """
    :type entity: copr.client_v2.entities.ProjectChrootEntity
    :type handle: copr.client_v2.handlers.ProjectChrootHandle
    """

    _schema = ProjectChrootSchema(**kwargs)

    def __init__(self, entity, handle, project, **kwargs):
        super(ProjectChroot, self).__init__(entity=entity, handle=handle, **kwargs)
        self._entity = entity
        self._handle = handle
        self._project = project

    @classmethod
    def from_response(cls, handle, data_dict, project, response=None, options=None):
        links = Link.from_dict(data_dict["_links"])
        entity = ProjectChrootEntity.from_dict(data_dict["chroot"])
        return cls(entity=entity, handle=handle, project=project,
                   response=response, links=links, options=options)

    def disable(self):
        """ Disables chroot for the bound project

        :rtype: :py:class:`.OperationResult`
        """
        return self._handle.disable(self._project, self.name)

    def update(self):
        """ Updates chroot with the current entity state

        :rtype: :py:class:`.OperationResult`
        """
        return self._handle.update(self._project, self._entity)


class MockChroot(IndividualResource):
    """
    :type entity: copr.client_v2.entities.MockChrootEntity
    :type handle: copr.client_v2.handlers.MockChrootHandle
    """

    _schema = MockChrootSchema(**kwargs)

    def __init__(self, entity, handle, **kwargs):
        super(MockChroot, self).__init__(
            entity=entity,
            handle=handle,
            **kwargs
        )

    @classmethod
    def from_response(cls, handle, data_dict, response=None, options=None):
        links = Link.from_dict(data_dict["_links"])
        entity = MockChrootEntity.from_dict(data_dict["chroot"])
        return cls(entity=entity, handle=handle,
                   response=response, links=links, options=options)


class OperationResult(IndividualResource):
    """ Fake resource to represent results of the requested operation

    """
    def __init__(self, handle, response=None, entity=None, options=None, expected_status=200):
        super(OperationResult, self).__init__(handle=handle, response=response, entity=entity, options=options)
        self._expected_status = expected_status

    @property
    def new_location(self):
        """ Contains an url to the new location produced by an operation
        If operation doesn't produce a new location would contain None

        :rtype: str
        """
        # todo: Create sub-class for results which contains `new_location`

        if self._response and \
                self._response.headers and \
                "location" in self._response.headers:
            return self._response.headers["location"]

        return None

    def is_successful(self):
        """ Performs check if status code is equal to the expected value
        of particular request.

        :rtype: bool
        """
        if self._response and self._response.status_code == self._expected_status:
            return True
        else:
            return False

    def __unicode__(self):
        out = u"<Result: "
        if self._response:
            out += u" status: {0}".format(self._response.status_code)
        out += u">"

        return out


class CollectionResource(Iterable, UnicodeMixin):
    """
    :type handle: client_v2.handlers.AbstractHandle or None
    :type response: ResponseWrapper
    :type links: (dict of (str, Link)) or None
    """

    def __init__(self, handle=None, response=None, links=None, individuals=None, options=None):
        self._handle = handle
        self._response = response
        self._links = links
        self._options = options or dict()
        self._individuals = individuals

    def get_href_by_name(self, name):
        """
        :type name: str
        """
        return self._links[name].href

    def next_page(self):
        limit = self._options.get("limit", 100)
        offset = self._options.get("offset", 0)

        offset += limit
        params = {}
        params.update(self._options)
        params["limit"] = limit
        params["offset"] = offset

        return self._handle.get_list(self, **params)

    def __iter__(self):
        """
        :rtype: Iterable[IndividualResource]
        """
        return iter(self._individuals)

    def __len__(self):
        if self._individuals is None:
            raise RuntimeError(u"Collection resource is missing self._individuals")

        return len(self._individuals)

    def __getitem__(self, item):
        if self._individuals is None:
            raise RuntimeError(u"Collection resource is missing self._individuals")

        return self._individuals[item]

    def __unicode__(self):
        out = u"<{}: [".format(self.__class__.__name__)
        out += u", ".join([str(x) for x in self])
        out += u"]>"
        return out

    @classmethod
    def from_response(cls, handle, response, options):
        raise NotImplementedError


def _construct_collection(
        resource_class, handle, response,
        individuals, options=None, **kwargs):

    """ Helper to avoid code repetition

    :type resource_class: CollectionResource
    :param handle: AbstractHandle
    :param response: ResponseWrapper
    :param options: dict with query options
    :param individuals: individuals

    :param kwargs: additional parameters for constructor

    :rtype: CollectionResource
    """
    return resource_class(
        handle,
        response=response,
        links=Link.from_dict(response.json["_links"]),
        individuals=individuals,
        options=options,
        **kwargs
    )


class ProjectList(CollectionResource):
    """
    :type handle: copr.client_v2.handlers.ProjectHandle
    """

    def __init__(self, handle, **kwargs):
        super(ProjectList, self).__init__(**kwargs)
        self._handle = handle

    def next_page(self):
        """
        Retrieves next chunk of the Project list for the same query options

        :rtype: :py:class:`.ProjectList`
        """
        return super(ProjectList, self).next_page()

    @property
    def projects(self):
        """ :rtype: list of :py:class:`~.resources.Project` """
        return self._individuals

    @classmethod
    def from_response(cls, handle, response, options):
        individuals = [
            Project.from_response(
                handle=handle,
                data_dict=dict_part,
            )
            for dict_part in response.json["projects"]
        ]
        return _construct_collection(
            cls, handle, response=response,
            individuals=individuals, options=options
        )


class BuildList(CollectionResource):
    """
    :type handle: copr.client_v2.handler.BuildHandle
    """
    def __init__(self, handle, **kwargs):
        super(BuildList, self).__init__(**kwargs)
        self._handle = handle

    def next_page(self):
        """
        Retrieves next chunk of the Build list for the same query options

        :rtype: :py:class:`.BuildList`
        """
        return super(BuildList, self).next_page()

    @property
    def builds(self):
        """
        :rtype: :py:class:`.BuildList`
        """
        return self._individuals

    @classmethod
    def from_response(cls, handle, response, options):
        individuals = [
            Build.from_response(
                handle=handle,
                data_dict=dict_part,
            )
            for dict_part in response.json["builds"]
        ]
        return _construct_collection(
            cls, handle, response=response,
            individuals=individuals, options=options
        )


class ProjectChrootList(CollectionResource):
    """
    List of the :class:`~.ProjectChroot` in the one Project.

    :type handle: copr.client_v2.handlers.ProjectChrootHandle
    """

    def __init__(self, handle, project, **kwargs):
        super(ProjectChrootList, self).__init__(**kwargs)
        self._handle = handle
        self._project = project

    @property
    def chroots(self):
        """
        :rtype: list of :py:class:`~.resources.ProjectChroot`
        """
        return self._individuals

    def enable(self, name):
        """
        Enables mock chroot for the current project

        :rtype: :py:class:`~.OperationResult`
        """
        return self._handle.enable(self._project, name)

    @classmethod
    def from_response(cls, handle, response, project):
        individuals = [
            ProjectChroot.from_response(
                handle=handle,
                data_dict=dict_part,
                project=project
            )
            for dict_part in response.json["chroots"]
        ]

        return _construct_collection(
            cls, handle, response=response,
            individuals=individuals, project=project
        )


class MockChrootList(CollectionResource):
    """
    List of the mock chroots supported by the service

    :type handle: copr.client_v2.handlers.MockChrootHandle
    """

    def __init__(self, handle, **kwargs):
        super(MockChrootList, self).__init__(**kwargs)
        self._handle = handle

    @property
    def chroots(self):
        """
        :rtype: list of :py:class:`~.resources.MockChroot`
        """
        return self._individuals

    @classmethod
    def from_response(cls, handle, response, options):
        individuals = [
            MockChroot.from_response(
                handle=handle,
                data_dict=dict_part,
            )
            for dict_part in response.json["chroots"]
        ]
        return _construct_collection(
            cls, handle, response=response,
            options=options, individuals=individuals
        )


class BuildTaskList(CollectionResource):
    """
    List of build tasks

    :type handle: copr.client_v2.handlers.BuildTaskHandle
    """

    def __init__(self, handle, **kwargs):
        super(BuildTaskList, self).__init__(**kwargs)
        self._handle = handle

    @property
    def build_tasks(self):
        """
        :rtype: list of :py:class:`~.resources.BuildTask`
        """
        return self._individuals

    @classmethod
    def from_response(cls, handle, response, options):
        individuals = [
            BuildTask.from_response(
                handle=handle,
                data_dict=dict_part
            )
            for dict_part in response.json["build_tasks"]
        ]
        return _construct_collection(
            cls, handle, response=response,
            options=options, individuals=individuals
        )



# coding: utf-8
import six
from copr.client_v2.handlers import ProjectHandle, ProjectChrootHandle

if six.PY3:
    from unittest import mock
    from unittest.mock import MagicMock
else:
    import mock
    from mock import MagicMock

import pytest

from copr.client_v2.client import CoprClient


class TestClientV2(object):
    def setup_method(self, method):
        self.login = u"foo"
        self.token = u"bar"
        self.root_url = "http://example.com/abcd"

        self.nc = MagicMock()

        self.root_json = {
            "_links": {
                "mock_chroots": {
                    "href": "/api_2/mock_chroots"
                },
                "self": {
                    "href": "/api_2/"
                },
                "projects": {
                    "href": "/api_2/projects"
                },
                "builds": {
                    "href": "/api_2/builds"
                },
                "build_tasks": {
                    "href": "/api_2/build_tasks"
                }
            }
        }
        self.response = MagicMock()

    @pytest.fixture
    def tc(self):
        # configures test client
        response = MagicMock()
        response.json = self.root_json
        self.nc.request.return_value = response

        client = CoprClient(self.nc, self.root_url, no_config=True)
        client.post_init()

        return client

    def test_create_from_empty_params(self):
        with mock.patch('copr.client_v2.client.NetClient') as handle:
            self.response.json = self.root_json
            handle.return_value.request.return_value = self.response
            CoprClient.create_from_params()

            assert handle.called
            assert handle.call_args == mock.call(None, None)

    def test_create_from_params(self):
        with mock.patch('copr.client_v2.client.NetClient') as handle:
            self.response.json = self.root_json
            handle.return_value.request.return_value = self.response
            CoprClient.create_from_params(
                self.root_url, self.login, self.token)

            assert handle.called
            assert handle.call_args == mock.call(self.login, self.token)

    # todo: def test_create_from_file_config

    def test_post_init(self):
        tc = CoprClient(self.nc, self.root_url, no_config=True)
        # import ipdb; ipdb.set_trace()

        self.response.json = self.root_json
        self.nc.request.return_value = self.response

        tc.post_init()
        assert tc._post_init_done

        assert isinstance(tc.projects, ProjectHandle)
        assert isinstance(tc.project_chroots, ProjectChrootHandle)

    def test_fixture_tc(self, tc):
        pass



# pylint: disable=R0903
"""
Wrappers for Copr api response.
Response classes provide convenient representation of received data
and offer actions based on response content.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

import weakref

#
# TODO:  Add Response for collections?
# TODO:  what about copr_project-> chroot_list, build_list, repos_list
#
from ..util import UnicodeMixin


class CoprResponse(UnicodeMixin):
    """
        Wrapper for Copr api responses

        :ivar handle: handle object which provide shortcuts
            based on request and/or response data
            (:py:class:`~.responses.BaseHandle` and its derivatives)
        :ivar dict data: json structure from Copr api
    """

    def __init__(self, client, method, data,
                 request_kwargs=None, parsers=None):
        """
            :param method: what method was used to query api
            :param data: json structure returned from Copr

            :param parsers: list of parser to process response data
            :type parsers: list of client.parsers.IParser
        """
        self.method = method
        self.data = data
        self.request_kwargs = request_kwargs
        self.parsers = parsers or []
        self.client = client
        self._parsed_data = {}
        self.handle = None

    def __getattr__(self, item):
        if self.data is None:
            raise RuntimeError("No response data to access sub-item")

        if item in self._parsed_data:
            return self._parsed_data[item]
        else:
            for parser in self.parsers:
                if item in parser.provided_fields:
                    value = parser.parse(self.data, item,
                                         client=self.client,
                                         request_kwargs=self.request_kwargs)
                    self._parsed_data[item] = value
                    return value
            raise KeyError(str(item))

    def __unicode__(self):
        return str(self.data)


class BaseHandle(object):
    """
    Handles provide convenient shortcut methods. Useful methods
    provided by derived classes.

    Example::

        response = client.create_project("copr")
        response.handle # <-- ProjectHandle object
        print(response.handle.get_project_details().data)

    """

    def __init__(self, client, username=None, response=None, **kwargs):
        """
            :param client: client instance used for request
            :type client: CoprClient
        """
        self.client = client
        self.username = username

        if response:
            self.response = weakref.proxy(response)
        else:
            self.response = None


class ProjectHandle(BaseHandle):
    """
        Handle to deal with a single Copr project
    """

    def __init__(self, client, projectname, *args, **kwargs):
        """
            :param client: client instance used for request
            :type client: CoprClient

            :param projectname: Copr project name
        """
        super(ProjectHandle, self).__init__(client, *args, **kwargs)
        self.projectname = projectname

    def get_project_details(self):
        """
            Shortcut to :meth:`~.client.CoprClient.get_project_details`
        """
        return self.client.get_project_details(
            self.projectname, username=self.username)

    def modify_project(self, **kwargs):
        """
            Shortcut to :meth:`~.client.CoprClient.modify_project`
        """
        return self.client.modify_project(
            self.projectname, username=self.username, **kwargs
        )

    def delete_project(self):
        """
            Shortcut to :meth:`~.client.CoprClient.delete_project`
        """
        return self.client.delete_project(
            self.projectname, username=self.username)

    def create_new_build(self, src_pkgs, chroots=None):
        """
            Shortcut to :meth"`~.client.CoprClient.create_new_build'
        """
        return self.client.create_new_build(self.projectname,
                                            src_pkgs, chroots)


class BuildHandle(BaseHandle):
    """
        Handle to deal with a single build
    """

    def __init__(self, client, build_id, *args, **kwargs):
        """
            :param client: client instance used for request
            :type client: CoprClient
        """
        super(BuildHandle, self).__init__(client, *args, **kwargs)
        self.build_id = build_id

        self.projectname = kwargs.get("projectname", None)

    @property
    def project_handle(self):
        """
            Shortcut for :py:class:`.responses.ProjectHandle`
        """
        if not self.projectname:
            raise Exception("Project name for build {0} is unknown".
                            format(self.build_id))
        if not self.username:
            raise Exception("Project owner for build {0} is unknown".
                            format(self.build_id))

        return ProjectHandle(client=self.client,
                             username=self.username,
                             projectname=self.projectname)

    def get_build_details(self):
        """
            Shortcut to :meth:`~.client.CoprClient.get_build_details`
        """
        return self.client.get_build_details(self.build_id,
                                             username=self.username,
                                             projectname=self.projectname)

    def cancel_build(self):
        """
            Shortcut to :meth:`~.client.CoprClient.cancel_build`
        """
        return self.client.cancel_build(self.build_id,
                                        username=self.username,
                                        projectname=self.projectname)


class ProjectChrootHandle(BaseHandle):
    """
        Handle to deal with a single project chroot
    """
    def __init__(self, client, chrootname, *args, **kwargs):
        super(ProjectChrootHandle, self).__init__(client, *args, **kwargs)

        self.chrootname = chrootname
        self.projectname = kwargs.get("projectname", None)

    @property
    def project_handle(self):
        if not self.projectname:
            raise Exception("Project name  is unknown")
        if not self.username:
            raise Exception("Project owner for build {0} is unknown")

        return ProjectHandle(client=self.client,
                             username=self.username,
                             projectname=self.projectname)

    def get_project_chroot_details(self):
        """
        Shortcut to :meth:`~.client.CoprClient.get_project_chroot_details`
        """
        return self.client.get_project_chroot_details(
            chrootname=self.chrootname,
            projectname=self.projectname,
            username=self.username)

    def modify_project_chroot_details(self, pkgs=None):
        """
        Shortcut to :meth:`~.client.CoprClient.modify_project_chroot_details`
        """
        return self.client.modify_project_chroot_details(
            self.projectname, self.chrootname, pkgs=pkgs)


class PackageHandle(BaseHandle):
    def __init__(self, client, ownername, projectname, name, *args, **kwargs):
        super(PackageHandle, self).__init__(client, *args, **kwargs)

        self.ownername = ownername
        self.projectname = projectname
        self.name = name


############################################################

class ProjectWrapper(UnicodeMixin):
    """
        Helper class to represent project objects

        ``__str__`` overridden to produces pretty formatted representation

        :ivar handle: :py:class:`.responses.ProjectHandle`

        :ivar username: project owner
        :ivar projectname: project names
    """

    def __init__(self, client, username, projectname,
                 description=None, instructions=None,
                 yum_repos=None, additional_repos=None):
        self.username = username
        self.projectname = projectname
        self.description = description
        self.instructions = instructions
        self.yum_repos = yum_repos or {}
        self.additional_repos = additional_repos or {}

        self.handle = ProjectHandle(client=client, projectname=projectname,
                                    username=username, response=None)

    def __unicode__(self):
        out = list()
        out.append(u"Name: {0}".format(self.projectname))
        out.append(u"  Description: {0}".format(self.description))

        if self.yum_repos:
            out.append(u"  Yum repo(s):")
            for k in sorted(self.yum_repos.keys()):
                out.append(u"    {0}: {1}".format(k, self.yum_repos[k]))
        if self.additional_repos:
            out.append(u"  Additional repo: {0}".format(self.additional_repos))
        if self.instructions:
            out.append(u"  Instructions: {0}".format(self.instructions))

        out.append(u"")
        return u"\n".join(out)


class BuildWrapper(object):
    """
        Helper class to represent build objects

        :ivar handle: :py:class:`.responses.BuildHandle`

        :ivar username: project owner
        :ivar projectname: project names
        :ivar int build_id: build identifier
    """

    def __init__(self, client, username, projectname, build_id,
                 status=None):
        self.username = username
        self.projectname = projectname
        self.build_id = build_id

        self.status = status

        self.handle = BuildHandle(client=client, build_id=build_id,
                                  projectname=projectname,
                                  username=username, response=None)


class ProjectChrootWrapper(object):
    """
        Helper class to represent project chroot objects

        :ivar handle: :py:class:`.responses.ProjectChrootHandle`

        :ivar username: project owner
        :ivar projectname: project names
        :ivar chrootname: chroot name
    """

    def __init__(self, client, username, projectname, chrootname,
                 repo_url=None):
        self.username = username
        self.projectname = projectname
        self.chrootname = chrootname

        self.repo_url = repo_url

        self.handle = ProjectChrootHandle(
            client=client, chrootname=chrootname,
            projectname=projectname,
            username=username, response=None
        )


class PackageWrapper(UnicodeMixin):
    def __init__(self, client, ownername, projectname, **data):
        self.client = client
        self.ownername = ownername
        self.projectname = projectname
        self.data = data
        self.handle = PackageHandle(
            client=self.client,
            ownername=self.ownername, projectname=self.projectname,
            copr_id=self.copr_id, enable_net=self.enable_net, name=self.name,
            id=self.id, source_json=self.source_json,
            source_type=self.source_type, webhook_rebuild=self.webhook_rebuild
        )

    def __getattr__(self, item):
        try:
            return self.data[item]
        except KeyError as e:
            raise AttributeError()

    def for_json(self):
        return self.data


class CoprChrootWrapper(UnicodeMixin):
    def __init__(self, client, ownername, projectname, **data):
        self.client = client
        self.ownername = ownername
        self.projectname = projectname
        self.data = data
        self.handle = None

    def __getattr__(self, item):
        try:
            return self.data[item]
        except KeyError as e:
            raise AttributeError()

    def for_json(self):
        return self.data



import os
from ..helpers import config_from_file
from ..requests import Request, munchify
from ..helpers import for_all_methods, bind_proxy


@for_all_methods(bind_proxy)
class BaseProxy(object):
    """
    Parent class for all other proxies
    """

    def __init__(self, config):
        self.config = config

    @classmethod
    def create_from_config_file(cls, path=None):
        config = config_from_file(path)
        return cls(config)

    @property
    def api_base_url(self):
        return os.path.join(self.config["copr_url"], "api_3", "")

    @property
    def auth(self):
        return self.config["login"], self.config["token"]

    def home(self):
        """
        Call the Copr APIv3 base URL

        :return: Munch
        """
        endpoint = ""
        request = Request(endpoint, api_base_url=self.api_base_url)
        response = request.send()
        return munchify(response)

    def auth_check(self):
        """
        Call an endpoint protected by login to check whether the user auth key is valid

        :return: Munch
        """
        endpoint = "/auth-check"
        request = Request(endpoint, api_base_url=self.api_base_url, auth=self.auth)
        response = request.send()
        return munchify(response)



# coding: utf-8
import six
import json
from copr.client_v2.net_client import ResponseWrapper

if six.PY3:
    from unittest.mock import MagicMock
else:
    from mock import MagicMock

import pytest

from copr.client_v2.handlers import ProjectHandle


class TestHandleBase(object):
    def setup_method(self, method):
        self.nc = MagicMock()
        self.client = MagicMock()

        self.root_url = "http://example.com"
        self.response = MagicMock()

        self.root_json = {
            "_links": {
                "mock_chroots": {
                    "href": "/api_2/mock_chroots"
                },
                "self": {
                    "href": "/api_2/"
                },
                "projects": {
                    "href": "/api_2/projects"
                },
                "builds": {
                    "href": "/api_2/builds"
                },
                "build_tasks": {
                    "href": "/api_2/build_tasks"
                }
            }
        }

    def get_href(self, name):
        return self.root_json["_links"][name]["href"]

    @pytest.fixture
    def project_handle(self):
        return ProjectHandle(self.client, self.nc, self.root_url,
                             self.get_href("projects"))

    def make_response(self, json_string, status=200, headers=None):
        response = MagicMock()
        response.status_code = status
        response.headers = headers or dict()
        response.content = json_string
        return ResponseWrapper(response)


class TestProjectHandle(TestHandleBase):
    project_1 = """{
        "project": {
            "description": "A simple KDE respin",
            "disable_createrepo": false,
            "repos": [],
            "contact": null,
            "owner": "jmiahman",
            "build_enable_net": true,
            "instructions": "",
            "homepage": null,
            "id": 2482,
            "name": "Synergy-Linux"
        },
        "project_chroots": [
            {
                "chroot": {
                    "comps": null,
                    "comps_len": 0,
                    "buildroot_pkgs": [],
                    "name": "fedora-19-x86_64",
                    "comps_name": null
                },
                "_links": null
            }
        ],
        "project_builds": [
            {
                "_links": null,
                "build": {
                    "enable_net": true,
                    "source_metadata": {
                        "url": "http://miroslav.suchy.cz/copr/copr-ping-1-1.fc20.src.rpm"
                    },
                    "submitted_on": 1422379448,
                    "repos": [],
                    "results": "https://copr-be.cloud.fedoraproject.org/results/jmiahman/Synergy-Linux/",
                    "started_on": 1422379466,
                    "source_type": 1,
                    "state": "succeeded",
                    "source_json": "{\\"url\\": \\"http://dl.kororaproject.org/pub/korora/releases/21/source/korora-welcome-21.6-1.fc21.src.rpm\\"}",
                    "ended_on": 1422379584,
                    "timeout": 21600,
                    "pkg_version": "21.6-1.fc21",
                    "id": 69493,
                    "submitter": "asamalik"
                }
            }
        ],
        "_links": {
            "self": {
              "href": "/api_2/projects/2482?show_builds=True&show_chroots=True"
            },
            "chroots": {
              "href": "/api_2/projects/2482/chroots"
            },
            "builds": {
              "href": "/api_2/builds?project_id=2482"
            }
        }
    }"""
    project_1_id = 2482
    project_1_owner = "jmiahman"
    project_1_name = "Synergy-Linux"

    project_list_1 = """
{"_links": {"self": {"href": "/api_2/projects?limit=2"}},
"projects": [{"project": {"group": "copr", "name": "copr", "disable_createrepo": false, "repos": [],
"description": "Lightweight buildsystem -", "contact": null, "owner": "msuchy",
"build_enable_net": false, "homepage": null, "id": 1,
"instructions": "See https://pagure.io/copr/copr for more details."},
"_links": {"self": {"href": "/api_2/projects/1"},
"chroots": {"href": "/api_2/projects/1/chroots"},
"build_tasks": {"href": "/api_2/build_tasks?project_id=1"},
"builds": {"href": "/api_2/builds?project_id=1"}}},
{"project": {"name": "abrt", "disable_createrepo": false, "repos": [],
"description": "Experimental ABRT ",
"contact": null, "owner": "jfilak",
"build_enable_net": true, "homepage": null, "id": 9,
"instructions": "git branches- libreport : rhel6_upstream_ureport - abrt : rhel6_auto_reporting"},
"_links": {"self": {"href": "/api_2/projects/9"}, "chroots": {"href": "/api_2/projects/9/chroots"},
"build_tasks": {"href": "/api_2/build_tasks?project_id=9"}, "builds": {"href": "/api_2/builds?project_id=9"}}}]}
    """

    def test_get_one(self, project_handle):
        response = self.make_response(self.project_1)
        self.nc.request.return_value = response
        project = project_handle.get_one(
            self.project_1_id,
            #    show_builds=True, show_chroots=True
        )

        assert project.name == self.project_1_name
        assert project.owner == self.project_1_owner

    def test_get_list(self, project_handle):
        response = self.make_response(self.project_list_1)
        self.nc.request.return_value = response
        plist = project_handle.get_list()

        projects = plist.projects
        assert len(projects) == 2
        assert set([p.name for p in projects]) == set(["copr", "abrt"])

    def test_get_list_pass_options(self, project_handle):
        response = self.make_response(self.project_list_1)
        self.nc.request.return_value = response

        query_params = {
            "search_query": "Foo bar",
            "owner": "John Smith",
            "name": "void",
            "offset": 12,
            "limit": 5
        }
        project_handle.get_list(**query_params)
        ca = self.nc.request.call_args

        expected = query_params.copy()
        expected["group"] = None

        assert ca[0][0] == self.root_url + "/api_2/projects"
        assert ca[1]["query_params"] == expected

    @pytest.fixture
    def one_project(self, project_handle):
        response = self.make_response(self.project_1)
        self.nc.request.return_value = response
        one = project_handle.get_one(self.project_1_id)
        self.nc.reset_mock()
        return one

    def test_get_self(self, one_project):
        new_p = one_project.get_self()
        assert new_p.name == one_project.name
        assert new_p.owner == one_project.owner

    def test_update(self, one_project):
        new_value = "foo bar"
        one_project.instructions = new_value
        assert self.nc.request.called is False
        one_project.update()
        assert self.nc.request.called
        ca = self.nc.request.call_args
        assert ca[1]["method"] == "put"
        assert json.loads(ca[1]["data"])["instructions"] == new_value

    def test_delete(self, one_project):
        response = self.make_response("", status=204)
        self.nc.request.return_value = response
        res = one_project.delete()
        assert res.is_successful()
        assert self.nc.request.called
        ca = self.nc.request.call_args

        assert ca[1]["method"] == "delete"
        assert ca[0][0] == self.root_url + "/api_2/projects/{0}".format(self.project_1_id)



# -*- coding: UTF-8 -*-
# pylint: disable=W1202

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

import json
import sys
import os
import logging
import io
import warnings

import requests
import six

from six.moves import configparser
from requests_toolbelt.multipart.encoder import (MultipartEncoder,
                                                 MultipartEncoderMonitor)

# urlparse from six is not available on el7
# because it requires at least python-six-1.4.1
if sys.version_info[0] == 2:
    from urlparse import urlparse
else:
    from urllib.parse import urlparse

if sys.version_info < (2, 7):
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
else:
    from logging import NullHandler

log = logging.getLogger(__name__)
log.addHandler(NullHandler())

from ..exceptions import CoprConfigException, CoprNoConfException, \
    CoprRequestException, \
    CoprUnknownResponseException

from .responses import ProjectHandle, \
    CoprResponse, BuildHandle, BaseHandle, ProjectChrootHandle

from .parsers import fabric_simple_fields_parser, ProjectListParser, \
    CommonMsgErrorOutParser, NewBuildListParser, ProjectChrootsParser, \
    ProjectDetailsFieldsParser, PackageListParser, PackageParser, \
    BuildConfigParser, CoprChrootParser

from ..util import UnicodeMixin

# TODO: add deco to check that login/token are provided
# and  raise correct error
# """ "No configuration file '~/.config/copr' found. "
# "see documentation at /usr/share/doc/python-copr/ "
# """
# or
# """
# "No api login and/or api token are provided"
#    "See man copr-cli for more information")
# """


SOURCE_TYPE_SRPM_LINK = 'srpm_link'
SOURCE_TYPE_SRPM_UPLOAD = 'srpm_upload'
SOURCE_TYPE_GIT_AND_TITO = 'git_and_tito'
SOURCE_TYPE_MOCK_SCM = 'mock_scm'
SOURCE_TYPE_PYPI = 'pypi'
SOURCE_TYPE_RUBYGEMS = 'rubygems'
SOURCE_TYPE_SCM = 'scm'
SOURCE_TYPE_CUSTOM = 'custom'

class CoprClient(UnicodeMixin):
    """ Main interface to the copr service

    :ivar unicode username: username used by default for all requests
    :ivar unicode login: user login, used for identification
    :ivar unicode token: copr api token
    :ivar unicode copr_url: used as copr projects root

    Could be created:
        - directly
        - using static method :py:meth:`CoprClient.create_from_file_config`

    """

    def __init__(self, username=None, login=None, token=None, copr_url=None,
                 no_config=False):
        """
            :param unicode username: username used by default for all requests
            :param unicode login: user login, used for identification
            :param unicode token: copr api token
            :param unicode copr_url: used as copr projects root
            :param bool no_config: helper flag to indicate that no config was provided
        """

        self.token = token
        self.login = login
        self.username = username
        self.copr_url = copr_url or "https://copr.fedorainfracloud.org/"

        self.no_config = no_config
        warnings.warn("You are using Copr's deprecated APIv1. "
                      "Please migrate to APIv3. "
                      "See https://fedora-copr.github.io/posts/EOL-APIv1-APIv2",
                      DeprecationWarning)

    def __unicode__(self):
        return (
            u"<Copr client. username: {0}, api url: {1}, login presents: {2}, token presents: {3}>"
            .format(self.username, self.api_url, bool(self.login), bool(self.token))
        )

    @property
    def api_url(self):
        """
            Url to API endpoint
        """
        return "{0}/api".format(self.copr_url)

    @staticmethod
    def create_from_file_config(filepath=None, ignore_error=False):
        """
        Creates Copr client using the information from the config file.

        :param filepath: specifies config location,
            default: "~/.config/copr"
        :type filepath: `str`
        :param bool ignore_error: When true creates default Client
            without credentionals

        :rtype: :py:class:`~.client.CoprClient`

        """

        raw_config = configparser.ConfigParser()
        if not filepath:
            filepath = os.path.join(os.path.expanduser("~"), ".config", "copr")
        config = {}

        try:
            exists = raw_config.read(filepath)
        except configparser.Error as e:
            raise CoprConfigException()

        if not exists:
            log.warning(
                "No configuration file '~/.config/copr' found. "
                "See man copr-cli for more information")
            config["no_config"] = True
            if not ignore_error:
                raise CoprNoConfException()
        else:
            try:
                for field in ["username", "login", "token", "copr_url"]:
                    if six.PY3:
                        config[field] = raw_config["copr-cli"].get(field, None)
                    else:
                        config[field] = raw_config.get("copr-cli", field, None)

            except configparser.Error as err:
                if not ignore_error:
                    raise CoprConfigException(
                        "Bad configuration file: {0}".format(err))
        return CoprClient(**config)

    def new_webhook_secret(self, coprname, ownername=None):
        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/{3}/".format(
            self.api_url, ownername, coprname, 'new_webhook_secret')

        data = self._fetch(url, method="post")

        response = CoprResponse(
            client=self,
            method="new_webhook_secret",
            data=data,
            parsers=[
                fabric_simple_fields_parser(["status", "output", "error", "message"]),
            ]
        )
        return response

    def authentication_check(self):
        url = "{0}/auth_check/".format(self.api_url)

        try:
            kwargs = {}
            kwargs["auth"] = (self.login, self.token)

            response = requests.request(
                method="POST",
                url=url,
                **kwargs
            )

            log.debug("raw response: {0}".format(response.text))

        except requests.ConnectionError as e:
            log.error(e)
            raise CoprRequestException("Connection error POST {0}".format(url))

        if not response.status_code in [200, 404]:
            try:
                output = json.loads(response.text)
            except ValueError:
                raise CoprUnknownResponseException(
                    "Unknown response from the server. Code: {0}, raw response:"
                    " \n {1}".format(response.status_code, response.text))
            raise CoprRequestException(output["error"])


    def _fetch(self, url, data=None, username=None, method=None,
               skip_auth=False, on_error_response=None, headers=None, params=None):
        """ Fetches data from server,
        checks response and raises a CoprRequestException with nice error message
        or CoprUnknownResponseException in case of some some error. \n
        Otherwise return unpacked json object.

            :param url: formed url to fetch
            :param data: [optional] serialised data to send
            :param skip_auth: [optional] don't send auth credentials
            :param username: [optional] use alternative username
            :param on_error_response: [optional] function to handle responses
                with bad status code
            :param headers: [optional] custom request headers
            :param params: [optional] data for GET requests

            :return: deserialized response
            :rtype: dict
        """

        if method is None:
            method = "get"

        if not username:
            username = self.username

        log.debug("Fetching url: {0}, for login: {1}".format(url, self.login))
        kwargs = {}
        if not skip_auth:
            kwargs["auth"] = (self.login, self.token)
        if data is not None:
            if type(data) not in [MultipartEncoder, MultipartEncoderMonitor]:
                data["username"] = username
            kwargs["data"] = data
        if headers is not None:
            kwargs["headers"] = headers
        if params is not None:
            kwargs["params"] = params

        if method not in ["get", "post", "head", "delete", "put"]:
            raise Exception("Method {0} not allowed".format(method))

        try:
            response = requests.request(
                method=method.upper(),
                url=url,
                **kwargs
            )
            log.debug("raw response: {0}".format(response.text))
        except requests.ConnectionError as e:
            log.error(e)
            raise CoprRequestException("Connection error {0} {1}".format(method.upper(), url))

        if "<title>Sign in Copr</title>" in response.text:
            raise CoprRequestException("Invalid API token\n")

        if response.status_code > 299 and on_error_response is not None:
            return on_error_response(response)

        if response.status_code == 404:
            log.error("Bad request, URL not found: {0}".
                      format(url))
        elif 400 <= response.status_code < 500:
            log.error("Bad request, raw response body: {0}".
                      format(response.text))
        elif response.status_code >= 500:
            log.error("Server error, raw response body: {0}".
                      format(response.text))

        try:
            output = json.loads(response.text)
        except ValueError:
            raise CoprUnknownResponseException(
                "Unknown response from the server. Code: {0}, raw response:"
                " \n {1}".format(response.status_code, response.text))
        if response.status_code != 200:
            raise CoprRequestException(output["error"])

        if output is None:
            raise CoprUnknownResponseException("No response from the server.")
        return output

    #########################################################
    ###                    Build actions                  ###
    #########################################################

    def get_build_details(self, build_id, projectname=None, username=None):
        """ Returns build details.

            :param build_id: Build identifier
            :type build_id: int

            :param projectname: [optional] Copr project name
            :param username: [optional] Copr project owner

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **handle:** :py:class:`~.responses.BuildHandle`
                - text fields: "project", "owner", "status", "results",
                "submitted_on", "started_on", "ended_on",
                "built_pkgs", "src_pkg", "src_version"


        """

        url = "{0}/coprs/build/{1}/".format(
            self.api_url, build_id)

        data = self._fetch(url, skip_auth=True)
        response = CoprResponse(
            client=self,
            method="get_build_details",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
                fabric_simple_fields_parser(
                    [
                        "project", "owner", "status", "results", "results_by_chroot",
                        "submitted_on", "started_on", "ended_on",
                        "built_pkgs", "src_pkg", "src_version",
                    ],  # TODO: convert unix time
                    "BuildDetailsParser"
                )
            ]
        )
        response.handle = BuildHandle(
            self, response=response, build_id=build_id,
            projectname=getattr(response, "project", projectname),
            username=getattr(response, "owner", username)
        )
        return response

    def cancel_build(self, build_id, projectname=None, username=None):
        """ Cancels build.
            Auth required.
            If build can't be canceled do nothing.

            :param build_id: Build identifier
            :type build_id: int

            :param projectname: [optional] Copr project name
            :param username: [optional] Copr project owner

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **handle:** :py:class:`~.responses.BuildHandle`
                - text fields: "status"
        """

        url = "{0}/coprs/cancel_build/{1}/".format(
            self.api_url, build_id)

        data = self._fetch(url, skip_auth=False, method='post')
        response = CoprResponse(
            client=self,
            method="cancel_build",
            data=data,
            parsers=[
                fabric_simple_fields_parser(["status", "output", "error"]),
            ]
        )
        response.handle = BuildHandle(
            self, response=response, build_id=build_id,
            projectname=projectname, username=username
        )
        return response

    def delete_build(self, build_id):
        """ Deletes build.
            Auth required.
            If build can't be deleted, return an error.

            :param build_id: Build identifier
            :type build_id: int

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **handle:** :py:class:`~.responses.BuildHandle`
                - text fields: "status"
        """

        url = "{0}/coprs/delete_build/{1}/".format(self.api_url, build_id)

        data = self._fetch(url, skip_auth=False, method='post')
        response = CoprResponse(
            client=self,
            method="delete_build",
            data=data,
            parsers=[
                fabric_simple_fields_parser(["status", "output", "error"]),
            ]
        )
        return response

    def create_new_build(self, projectname, pkgs, username=None,
                         timeout=None, memory=None, chroots=None,
                         background=False, progress_callback=None):
        """ Creates new build

            :param projectname: name of Copr project (without user namespace)
            :param pkgs: list of packages to include in build
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        data = {
            "memory_reqs": memory,
            "timeout": timeout
        }

        if urlparse(pkgs[0]).scheme != "":
            api_endpoint = "new_build"
            data["pkgs"] = " ".join(pkgs)
        else:
            try:
                api_endpoint = "new_build_upload"
                f = open(pkgs[0], "rb")
                data["pkgs"] = (os.path.basename(f.name), f, "application/x-rpm")
            except IOError as e:
                raise CoprRequestException(e)

        return self.process_creating_new_build(projectname, data, api_endpoint, username, chroots, background=background,
                                               progress_callback=progress_callback, multipart=True)

    def create_new_build_pypi(self, projectname, pypi_package_name, pypi_package_version=None,
                              spec_template="", python_versions=[3, 2], username=None, timeout=None,
                              memory=None, chroots=None, background=False, progress_callback=None):
        """ Creates new build from PyPI

            :param projectname: name of Copr project (without user namespace)
            :param pypi_package_name: PyPI package name
            :param pypi_package_vesion: [optional] PyPI package version (None means "latest")
            :param spec_template: [optional] what spec template to use
            :param python_versions: [optional] list of python versions to build for
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        data = {
            "memory_reqs": memory,
            "timeout": timeout,
            "pypi_package_name": pypi_package_name,
            "pypi_package_version": pypi_package_version,
            "spec_template": spec_template,
            "python_versions": [str(version) for version in python_versions],
        }
        api_endpoint = "new_build_pypi"
        return self.process_creating_new_build(projectname, data, api_endpoint, username,
                                               chroots, background=background)

    def create_new_build_tito(self, projectname, git_url, git_dir=None, git_branch=None, tito_test=None, username=None,
                              timeout=None, memory=None, chroots=None, background=False, progress_callback=None):
        """ Creates new build from PyPI

            :param projectname: name of Copr project (without user namespace)
            :param git_url: url to Git code which is able to build via Tito
            :param git_dir: [optional] path to directory containing .spec file
            :param git_branch: [optional] git branch
            :param tito_test: [optional] build the last commit instead of the last release tag
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        print('Deprecated method. Use generic "scm" methods instead.', file=sys.stderr)

        data = {
            "memory_reqs": memory,
            "timeout": timeout,
            "git_url": git_url,
            "git_directory": git_dir,  # @FIXME
            "git_branch": git_branch,
            "tito_test": tito_test,
        }
        api_endpoint = "new_build_tito"
        return self.process_creating_new_build(projectname, data, api_endpoint, username,
                                               chroots, background=background)

    def create_new_build_mock(self, projectname, scm_url, spec, scm_type="git", scm_branch=None, username=None,
                              timeout=None, memory=None, chroots=None, background=False, progress_callback=None):
        """ Creates new build from PyPI

            :param projectname: name of Copr project (without user namespace)
            :param scm_url: url to a project versioned by Git or SVN
            :param spec: relative path from SCM root to .spec file
            :param scm_type: possible values are "git" and "svn"
            :param scm_branch: [optional] Git or SVN branch
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        print('Deprecated method. Use generic "scm" methods instead.', file=sys.stderr)

        data = {
            "memory_reqs": memory,
            "timeout": timeout,
            "scm_type": scm_type,
            "scm_url": scm_url,
            "scm_branch": scm_branch,
            "spec": spec,
        }
        api_endpoint = "new_build_mock"
        return self.process_creating_new_build(projectname, data, api_endpoint, username,
                                               chroots, background=background)

    def create_new_build_scm(self, projectname, clone_url, committish='', subdirectory='', spec='',
                             scm_type='git', srpm_build_method='rpkg', username=None, timeout=None,
                             memory=None, chroots=None, background=False, progress_callback=None):
        """ Creates new build from SCM

            :param projectname: name of Copr project (without user namespace)
            :param clone_url: url to a project versioned by Git or SVN
            :param committish [optional]: name of a branch, tag, or a git hash
            :param subdirectory [optional]: repo subdirectory with package content
            :param spec [optional]: path to spec file, relative to 'subdirectory'
            :param scm_type [optional]: "git" or "svn"
            :param srpm_build_method [optional]: tool to build srpm with. One of:
                "rpkg", "tito", "tito_test", "make_srpm"
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        data = {
            "memory_reqs": memory,
            "timeout": timeout,
            "clone_url": clone_url,
            "committish": committish,
            "subdirectory": subdirectory,
            "spec": spec,
            "scm_type": scm_type,
            "srpm_build_method": srpm_build_method,
        }
        api_endpoint = "new_build_scm"
        return self.process_creating_new_build(projectname, data, api_endpoint, username,
                                               chroots, background=background)

    def create_new_build_rubygems(self, projectname, gem_name, username=None,
                              timeout=None, memory=None, chroots=None, background=False, progress_callback=None):
        """ Creates new build from RubyGems.org

            :param projectname: name of Copr project (without user namespace)
            :param gem_name: name of the gem located on rubygems.org
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        data = {
            "memory_reqs": memory,
            "timeout": timeout,
            "gem_name": gem_name,
        }
        api_endpoint = "new_build_rubygems"
        return self.process_creating_new_build(projectname, data, api_endpoint, username,
                                               chroots, background=background)


    def create_new_build_custom(self, projectname,
            script, script_chroot=None, script_builddeps=None,
            script_resultdir=None,
            username=None, timeout=None, memory=None, chroots=None,
            background=False, progress_callback=None):
        """ Creates new build with Custom source build method.

            :param projectname: name of Copr project (without user namespace)
            :param script: script to execute to generate sources
            :param script_chroot: [optional] what chroot to use to generate
                sources (defaults to fedora-latest-x86_64)
            :param script_builddeps: [optional] list of script's dependencies
            :param script_resultdir: [optional] where script generates results
                (relative to cwd)
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        data = {
            "memory_reqs": memory,
            "timeout": timeout,
            "script": script,
            "chroot": script_chroot,
            "builddeps": script_builddeps,
            "resultdir": script_resultdir,
        }

        api_endpoint = "new_build_custom"
        return self.process_creating_new_build(projectname, data, api_endpoint, username,
                                               chroots, background=background)


    def create_new_build_distgit(self, projectname, clone_url, branch=None, username=None,
                              timeout=None, memory=None, chroots=None, background=False, progress_callback=None):
        """ Creates new build from a dist-git repository

            :param projectname: name of Copr project (without user namespace)
            :param clone_url: url of the distgit repository to be cloned
            :param branch: [optional] branch in the repository to be checked out
            :param username: [optional] use alternative username
            :param timeout: [optional] build timeout
            :param memory: [optional] amount of required memory for build process
            :param chroots: [optional] build only with given chroots
            :param background: [optional] mark the build as a background job.
            :param progress_callback: [optional] a function that received a
            MultipartEncoderMonitor instance for each chunck of uploaded data

            :return: :py:class:`~.responses.CoprResponse` with additional fields:

                - **builds_list**: list of :py:class:`~.responses.BuildWrapper`
        """
        data = {
            "memory_reqs": memory,
            "timeout": timeout,
            "clone_url": clone_url,
            "branch": branch,
        }
        api_endpoint = "new_build_distgit"
        return self.process_creating_new_build(projectname, data, api_endpoint, username,
                                               chroots, background=background)

    def process_creating_new_build(self, projectname, data, api_endpoint, username=None, chroots=None,
                                   background=False, progress_callback=None, multipart=False):
        if not username:
            username = self.username
        data["username"] = self.username

        url = "{0}/coprs/{1}/{2}/{3}/".format(
            self.api_url, username, projectname, api_endpoint
        )

        if background:
            data["background"] = "y"

        for chroot in chroots or []:
            data[chroot] = "y"

        if not multipart:
            data = self._fetch(url, data, method="post")
        else:
            m = MultipartEncoder(data)
            callback = progress_callback or (lambda x: x)
            monit = MultipartEncoderMonitor(m, callback)
            data = self._fetch(url, monit, method="post",
                               headers={'Content-Type': monit.content_type})

        response = CoprResponse(
            client=self,
            method="cancel_build",
            data=data,
            request_kwargs={
                "projectname": projectname,
                "username": username
            },
            parsers=[
                CommonMsgErrorOutParser,
                NewBuildListParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname, username=username)

        return response


    #########################################################
    ###                   Package actions                 ###
    #########################################################

    def get_package_edit_url(self, ownername, projectname, package_name, source_type):
        return "{0}/coprs/{1}/{2}/package/{3}/edit/{4}/".format(
            self.api_url, ownername or self.username, projectname, package_name, source_type
        )

    def get_package_add_url(self, ownername, projectname, source_type):
        return "{0}/coprs/{1}/{2}/package/add/{3}/".format(
            self.api_url, ownername or self.username, projectname, source_type
        )

    def get_package_delete_url(self, ownername, projectname, package_name):
        return "{0}/coprs/{1}/{2}/package/{3}/delete/".format(
            self.api_url, ownername or self.username, projectname, package_name
        )

    def edit_package_tito(self, package_name, projectname, git_url, git_dir=None, git_branch=None, tito_test=None, ownername=None, webhook_rebuild=None):
        print('Deprecated method. Use generic "scm" methods instead.', file=sys.stderr)
        request_url = self.get_package_edit_url(ownername, projectname, package_name, SOURCE_TYPE_GIT_AND_TITO)
        data = {
            "package_name": package_name,
            "git_url": git_url,
            "git_directory": git_dir,
            "git_branch": git_branch,
            "tito_test": 'y' if tito_test else '', # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better
        }
        if webhook_rebuild != None:
            data['webhook_rebuild'] = 'y' if webhook_rebuild else '' # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better

        response = self.process_package_action(request_url, ownername, projectname, data)
        return response

    def add_package_tito(self, package_name, projectname, git_url, git_dir=None, git_branch=None, tito_test=None, ownername=None, webhook_rebuild=None):
        print('Deprecated method. Use generic "scm" methods instead.', file=sys.stderr)
        request_url = self.get_package_add_url(ownername, projectname, SOURCE_TYPE_GIT_AND_TITO)
        response = self.process_package_action(request_url, ownername, projectname, data={
            "package_name": package_name,
            "git_url": git_url,
            "git_directory": git_dir,
            "git_branch": git_branch,
            "tito_test": 'y' if tito_test else '', # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better
            "webhook_rebuild": 'y' if webhook_rebuild else '', # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better
        })
        return response

    def edit_package_pypi(self, package_name, projectname, pypi_package_name, pypi_package_version,
                          spec_template="", python_versions=[3, 2], ownername=None, webhook_rebuild=None):
        request_url = self.get_package_edit_url(ownername, projectname, package_name, SOURCE_TYPE_PYPI)
        data = {
            "package_name": package_name,
            "pypi_package_name": pypi_package_name,
            "pypi_package_version": pypi_package_version,
            "spec_template": spec_template,
            "python_versions": python_versions,
        }
        if webhook_rebuild != None:
            data['webhook_rebuild'] = 'y' if webhook_rebuild else '' # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better

        response = self.process_package_action(request_url, ownername, projectname, data)
        return response

    def add_package_pypi(self, package_name, projectname, pypi_package_name, pypi_package_version,
                         spec_template="", python_versions=[3, 2], ownername=None, webhook_rebuild=None):
        request_url = self.get_package_add_url(ownername, projectname, SOURCE_TYPE_PYPI)
        response = self.process_package_action(request_url, ownername, projectname, data={
            "package_name": package_name,
            "pypi_package_name": pypi_package_name,
            "pypi_package_version": pypi_package_version,
            "spec_template": spec_template,
            "python_versions": python_versions,
            "webhook_rebuild": 'y' if webhook_rebuild else '',
        })
        return response

    def edit_package_mockscm(self, package_name, projectname, scm_type, scm_url, scm_branch, spec, ownername=None, webhook_rebuild=None):
        print('Deprecated method. Use generic "scm" methods instead.', file=sys.stderr)
        request_url = self.get_package_edit_url(ownername, projectname, package_name, SOURCE_TYPE_MOCK_SCM)
        data = {
            "package_name": package_name,
            "scm_type": scm_type,
            "scm_url": scm_url,
            "scm_branch": scm_branch,
            "spec": spec,
        }
        if webhook_rebuild != None:
            data['webhook_rebuild'] = 'y' if webhook_rebuild else '' # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better

        response = self.process_package_action(request_url, ownername, projectname, data)
        return response

    def add_package_mockscm(self, package_name, projectname, scm_type, scm_url, scm_branch, spec, ownername=None, webhook_rebuild=None):
        print('Deprecated method. Use generic "scm" methods instead.', file=sys.stderr)
        request_url = self.get_package_add_url(ownername, projectname, SOURCE_TYPE_MOCK_SCM)
        response = self.process_package_action(request_url, ownername, projectname, data={
            "package_name": package_name,
            "scm_type": scm_type,
            "scm_url": scm_url,
            "scm_branch": scm_branch,
            "spec": spec,
            "webhook_rebuild": 'y' if webhook_rebuild else '',
        })
        return response

    def edit_package_scm(self, package_name, projectname, clone_url, committish='', subdirectory='', spec='',
                         scm_type='git', srpm_build_method='rpkg', ownername=None, webhook_rebuild=None):
        request_url = self.get_package_edit_url(ownername, projectname, package_name, SOURCE_TYPE_SCM)
        data = {
            "package_name": package_name,
            "clone_url": clone_url,
            "committish": committish,
            "subdirectory": subdirectory,
            "spec": spec,
            "scm_type": scm_type,
            "srpm_build_method": srpm_build_method,
        }
        if webhook_rebuild != None:
            data['webhook_rebuild'] = 'y' if webhook_rebuild else '' # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better

        response = self.process_package_action(request_url, ownername, projectname, data)
        return response

    def add_package_scm(self, package_name, projectname, clone_url, committish='', subdirectory='', spec='',
                        scm_type='git', srpm_build_method='rpkg', ownername=None, webhook_rebuild=None):
        request_url = self.get_package_add_url(ownername, projectname, SOURCE_TYPE_SCM)
        response = self.process_package_action(request_url, ownername, projectname, data={
            "package_name": package_name,
            "clone_url": clone_url,
            "committish": committish,
            "subdirectory": subdirectory,
            "spec": spec,
            "scm_type": scm_type,
            "srpm_build_method": srpm_build_method,
            "webhook_rebuild": 'y' if webhook_rebuild else '',
        })
        return response

    def edit_package_rubygems(self, package_name, projectname, gem_name, ownername=None, webhook_rebuild=None):
        request_url = self.get_package_edit_url(ownername, projectname, package_name, SOURCE_TYPE_RUBYGEMS)
        data = {
            "package_name": package_name,
            "gem_name": gem_name,
        }
        if webhook_rebuild != None:
            data['webhook_rebuild'] = 'y' if webhook_rebuild else '' # TODO: False/True gets converted to 'False'/'True' in FE, try to solve better

        response = self.process_package_action(request_url, ownername, projectname, data)
        return response

    def add_package_rubygems(self, package_name, projectname, gem_name, ownername=None, webhook_rebuild=None):
        request_url = self.get_package_add_url(ownername, projectname, SOURCE_TYPE_RUBYGEMS)
        response = self.process_package_action(request_url, ownername, projectname, data={
            "package_name": package_name,
            "gem_name": gem_name,
            "webhook_rebuild": 'y' if webhook_rebuild else '',
        })
        return response

    def edit_package_custom(self, package_name, projectname,
            script, script_chroot=None, script_builddeps=None,
            script_resultdir=None,
            ownername=None, webhook_rebuild=None):

        request_url = self.get_package_edit_url(ownername, projectname,
                package_name, SOURCE_TYPE_CUSTOM)

        data = {
            "package_name": package_name,
            "script": script,
            "builddeps": script_builddeps,
            "resultdir": script_resultdir,
            "chroot": script_chroot,
        }
        if webhook_rebuild != None:
            data['webhook_rebuild'] = 'y' if webhook_rebuild else ''

        response = self.process_package_action(request_url, ownername, projectname, data)
        return response

    def add_package_custom(self, package_name, projectname,
            script, script_chroot=None, script_builddeps=None,
            script_resultdir=None,
            ownername=None, webhook_rebuild=None):

        request_url = self.get_package_add_url(ownername, projectname,
                SOURCE_TYPE_CUSTOM)
        response = self.process_package_action(request_url, ownername,
                projectname, data={
                    "package_name": package_name,
                    "script": script,
                    "builddeps": script_builddeps,
                    "resultdir": script_resultdir,
                    "chroot": script_chroot,
                    "webhook_rebuild": 'y' if webhook_rebuild else '',
                },
        )
        return response

    def process_package_action(self, request_url, ownername, projectname, data, fetch_functor=None):
        if not ownername:
            ownername = self.username

        # @TODO refactor this hacky part
        # @TODO I want to have this function for various kind of actions, not only packages
        if fetch_functor:
            resp_data = fetch_functor(request_url, data, method="post")
        else:
            resp_data = self._fetch(request_url, data, method="post")

        response = CoprResponse(
            client=self,
            method="post",
            data=resp_data,
            request_kwargs={
                "projectname": projectname,
                "ownername": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname,
            username=ownername
        )

        return response

    def get_packages_list(self, projectname, with_latest_build=False, with_latest_succeeded_build=False, with_all_builds=False, ownername=None):
        """Returns list of packages for the given copr."""

        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/package/list/".format(
            self.api_url, ownername, projectname
        )

        resp_data = self._fetch(url, params={
            "with_latest_build": 'y' if with_latest_build else '',
            "with_latest_succeeded_build": 'y' if with_latest_succeeded_build else '',
            "with_all_builds": 'y' if with_all_builds else '',
        })
        response = CoprResponse(
            client=self,
            method="get",
            data=resp_data,
            request_kwargs={
                "projectname": projectname,
                "ownername": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
                PackageListParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname,
            username=ownername
        )

        return response

    def get_package(self, projectname, pkg_name, with_latest_build=False, with_latest_succeeded_build=False, with_all_builds=False, ownername=None):
        """Returns single package if pkg_name."""

        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/package/get/{3}/".format(
            self.api_url, ownername, projectname, pkg_name
        )

        resp_data = self._fetch(url, params={
            "with_latest_build": 'y' if with_latest_build else '',
            "with_latest_succeeded_build": 'y' if with_latest_succeeded_build else '',
            "with_all_builds": 'y' if with_all_builds else '',
        })
        response = CoprResponse(
            client=self,
            method="get",
            data=resp_data,
            request_kwargs={
                "projectname": projectname,
                "ownername": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
                PackageParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname, username=ownername
        )

        return response

    def delete_package(self, projectname, pkg_name, ownername=None):
        """Deletes the given package."""

        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/package/delete/{3}/".format(
            self.api_url, ownername, projectname, pkg_name
        )

        resp_data = self._fetch(url, method="post")
        response = CoprResponse(
            client=self,
            method="post",
            data=resp_data,
            request_kwargs={
                "projectname": projectname,
                "ownername": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname, username=ownername
        )

        return response

    def reset_package(self, projectname, pkg_name, ownername=None):
        """Resets default source of the given package."""

        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/package/reset/{3}/".format(
            self.api_url, ownername, projectname, pkg_name
        )

        resp_data = self._fetch(url, method="post")
        response = CoprResponse(
            client=self,
            method="post",
            data=resp_data,
            request_kwargs={
                "projectname": projectname,
                "ownername": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname, username=ownername
        )

        return response

    def build_package(self, projectname, pkg_name, ownername=None, chroots=None, timeout=None):
        """Builds the package from its default source."""

        if not ownername:
            ownername = self.username

        request_url = "{0}/coprs/{1}/{2}/package/build/{3}/".format(
            self.api_url, ownername, projectname, pkg_name
        )

        data = {}
        for chroot in chroots or []:
            data[chroot] = "y"
        if timeout:
            data["timeout"] = timeout

        resp_data = self._fetch(request_url, data, method="post")
        response = CoprResponse(
            client=self,
            method="post",
            data=resp_data,
            request_kwargs={
                "projectname": projectname,
                "username": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
                NewBuildListParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname, username=ownername
        )

        return response

    #########################################################
    ###                   Project actions                 ###
    #########################################################

    def get_project_details(self, projectname, username=None):
        """ Returns project details

            :param projectname: Copr projectname
            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - text fields: "description", "instructions", "last_modified",
                "name"

                - **chroots**: list of
                  :py:class:`~.responses.ProjectChrootWrapper`
        """
        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/detail/".format(
            self.api_url, username, projectname
        )

        data = self._fetch(url, skip_auth=True)
        # return ProjectDetailsResponse(self, response, projectname, username)

        response = CoprResponse(
            client=self,
            method="get_project_details",
            data=data,
            request_kwargs={
                "projectname": projectname,
                "username": username
            },
            parsers=[
                ProjectChrootsParser,
                ProjectDetailsFieldsParser,
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname,
                                        username=username)
        return response

    def delete_project(self, projectname, username=None):
        """ Deletes the entire project.
            Auth required.

            :param projectname: Copr projectname
            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - text fields: "message"
        """

        if not username:
            username = self.username
        url = "{0}/coprs/{1}/{2}/delete/".format(
            self.api_url, username, projectname
        )

        data = self._fetch(
            url, data={"verify": "yes"}, method="post")

        response = CoprResponse(
            client=self,
            method="delete_project",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname,
                                        username=username)
        return response

    def fork_project(self, source, projectname, username=None, confirm=False):
        """ Fork the project and builds in it
        Auth required.

        :param source: source Copr name or full_name
        :param projectname: destination Copr projectname
        :param username: [optional] use alternative username as owner of forked project
        :param confirm: [optional] need to pass True when forking into existing project

        :return: :py:class:`~.responses.CoprResponse`
        with additional fields:

        - text fields: "message"
        """
        if not username:
            username = self.username
        url = "{0}/coprs/{1}/fork/".format(
            self.api_url, source
        )

        post_data = {
            "name": projectname,
            "owner": username,
            "source": source,
        }
        if confirm:
            post_data["confirm"] = confirm
        data = self._fetch(url, data=post_data, method="post")

        response = CoprResponse(
            client=self,
            method="fork_project",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname,
                                        username=username)
        return response

    def create_project(
            self, username, projectname, chroots,
            description=None, instructions=None,
            repos=None, initial_pkgs=None, disable_createrepo=None,
            unlisted_on_hp=False, enable_net=True, persistent=False,
            auto_prune=True, use_bootstrap_container=None,
    ):
        """ Creates a new copr project
            Auth required.

            :param projectname: User or group name
            :param projectname: Copr project name
            :param chroots: List of target chroots
            :param description: [optional] Project description
            :param instructions: [optional] Instructions for end users
            :param disable_createrepo: [optional] disables automatic repo meta-data regeneration, "true"/"false" string
            :param unlisted_on_hp: [optional] Project will not be shown on COPR HP
            :param enable_net: [optional] If builder can access net for builds in this project
            :param persistent: [optional] If builds and the project are undeletable
            :param auto_prune: [optional] If backend auto-deletion script should be run for the project
            :param use_bootstrap_container: [optional] If mock bootstrap container is used to initialize the buildroot

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectHandle`
                - text fields: "message"
        """

        if not username:
            username = self.username

        url = "{0}/coprs/{1}/new/".format(
            self.api_url, username)

        if not chroots:
            raise Exception("You should provide chroots")

        if not isinstance(chroots, list):
            chroots = [chroots]

        if isinstance(repos, list):
            repos = " ".join(repos)

        if isinstance(initial_pkgs, list):
            initial_pkgs = " ".join(initial_pkgs)

        request_data = {
            "name": projectname,
            "repos": repos,
            "initial_pkgs": initial_pkgs,
            "description": description,
            "instructions": instructions,
            "disable_createrepo": disable_createrepo,
            "unlisted_on_hp": "y" if unlisted_on_hp else "",
            "build_enable_net": "y" if enable_net else "",
            "persistent": "y" if persistent else "",
            "auto_prune": "y" if auto_prune else "",
            "use_bootstrap_container": "y" if use_bootstrap_container else "",
        }
        for chroot in chroots:
            request_data[chroot] = "y"

        # TODO: def on bad_response()
        result_data = self._fetch(url, data=request_data, method="post")

        response = CoprResponse(
            client=self,
            method="create_project",
            data=result_data,
            parsers=[
                CommonMsgErrorOutParser,
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname)
        return response

    def modify_project(self, projectname, username=None,
                       description=None, instructions=None,
                       repos=None, disable_createrepo=None, unlisted_on_hp=None,
                       enable_net=None, auto_prune=None,
                       use_bootstrap_container=None, chroots=None):
        """ Modifies main project configuration.
            Auth required.

            :param projectname: Copr project name
            :param username: [optional] use alternative username
            :param description: [optional] project description
            :param instructions: [optional] instructions for end users
            :param repos: [optional] list of additional repos to be used during
                the build process
            :param repos: [optional] list of additional repos to be used during
            :param disable_createrepo: [optional] disables automatic repo meta-data regeneration
            :param unlisted_on_hp: [optional] Project will not be shown on COPR HP
            :param enable_net: [optional] If builder can access net for builds in this project
            :param auto_prune: [optional] If backend auto-deletion script should be run for the project
            :param use_bootstrap_container: [optional] If mock bootstrap container is used to initialize the buildroot
            :param chroots: [optional] list of chroots that should be enabled in the project. When not ``None``,
                selected chroots will be enabled while current chroots
                will not remain enabled if they are not specified.

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectHandle`
                - text fields: "buildroot_pkgs"
        """

        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/modify/".format(
            self.api_url, username, projectname
        )
        data = {}
        if description:
            data["description"] = description
        if instructions:
            data["instructions"] = instructions
        if repos:
            data["repos"] = repos
        if disable_createrepo:
            data["disable_createrepo"] = disable_createrepo
        if unlisted_on_hp != None:
            data["unlisted_on_hp"] = "y" if unlisted_on_hp else ""
        if enable_net != None:
            data["build_enable_net"] = "y" if enable_net else ""
        if auto_prune != None:
            data["auto_prune"] = "y" if auto_prune else ""
        if use_bootstrap_container != None:
            data["use_bootstrap_container"] = "y" if use_bootstrap_container else ""
        if chroots != None:
            data["chroots"] = " ".join(chroots)

        result_data = self._fetch(url, data=data, method="post")

        response = CoprResponse(
            client=self,
            method="modify_project",
            data=result_data,
            parsers=[
                CommonMsgErrorOutParser,
                ProjectDetailsFieldsParser,
                fabric_simple_fields_parser(["buildroot_pkgs"])
            ]
        )
        response.handle = ProjectHandle(client=self, response=response,
                                        projectname=projectname)
        return response

    def get_projects_list(self, username=None):
        """ Returns list of projects created by the user

            :param username: [optional] use alternative username


            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **projects_list**: list of
                  :py:class:`~.responses.ProjectWrapper`
        """
        if not username:
            username = self.username

        url = "{0}/coprs/{1}/".format(
            self.api_url, username)
        data = self._fetch(url)
        response = CoprResponse(
            client=self,
            method="get_projects_list",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
                ProjectListParser,
            ]
        )
        response.handle = BaseHandle(client=self, username=username,
                                     response=response)
        return response

    def edit_chroot(self, projectname, chrootname, ownername=None,
                      upload_comps=None, delete_comps=None, packages=None, repos=None):
        """ Edits chroot settings.
            Auth required.

            :param projectname: Copr project name
            :param chrootname: chroot name
            :param ownername: [optional] owner of the project
            :param upload_comps: file path to the comps.xml file
            :param delete_comps: True if comps.xml should be removed
            :param packages: buildroot packages for the chroot
            :param repos: buildroot additional repos

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectHandle`
                - text fields: "buildroot_pkgs"
        """

        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/chroot/edit/{3}/".format(
            self.api_url, ownername, projectname, chrootname
        )
        multipart = False
        headers = None
        data = {}
        if upload_comps:
            try:
                f = open(upload_comps, "rb")
                data["upload_comps"] = (os.path.basename(f.name), f, "application/text")
                multipart = True
            except IOError as e:
                raise CoprRequestException(e)
        if delete_comps != None:
            data["delete_comps"] = "y" if delete_comps else ""
        if packages != None:
            data["buildroot_pkgs"] = packages
        if repos != None:
            data["repos"] = repos

        if multipart:
            data = MultipartEncoder(data)
            headers={'Content-Type': data.content_type}

        result_data = self._fetch(url, data, method="post", headers=headers)

        response = CoprResponse(
            client=self,
            method="post",
            data=result_data,
            request_kwargs={
                "projectname": projectname,
                "ownername": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
                CoprChrootParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname,
            username=ownername
        )
        return response

    def get_chroot(self, projectname, ownername, chrootname=None):
        """Returns copr_chroot data"""

        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/chroot/get/{3}/".format(
            self.api_url, ownername, projectname, chrootname
        )
        resp_data = self._fetch(url)

        response = CoprResponse(
            client=self,
            method="get",
            data=resp_data,
            request_kwargs={
                "projectname": projectname,
                "ownername": ownername
            },
            parsers=[
                CommonMsgErrorOutParser,
                CoprChrootParser,
            ]
        )
        response.handle = BaseHandle(
            self, response=response,
            projectname=projectname,
            username=ownername
        )

        return response

    def get_project_chroot_details(self, projectname,
                                   chrootname, username=None):
        """ Returns details of chroot used in project

            :param projectname: Copr project name
            :param chrootname: chroot name

            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectChrootHandle`
                - text fields: "buildroot_pkgs"
        """
        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/detail/{3}/".format(
            self.api_url, username, projectname, chrootname
        )
        data = self._fetch(url, skip_auth=True)
        response = CoprResponse(
            client=self,
            method="get_project_chroot_details",
            data=data,
            parsers=[
                fabric_simple_fields_parser(
                    ["buildroot_pkgs", "output", "error"],
                    "BuildDetailsParser"
                )
            ]
        )
        response.handle = ProjectChrootHandle(
            client=self, chrootname=chrootname, username=username,
            projectname=projectname, response=response
        )
        return response

    def modify_project_chroot_details(self, projectname, chrootname,
                                      pkgs=None, username=None):
        """ @deprecated to edit_chroot
            Modifies chroot used in project

            :param projectname: Copr project name
            :param chrootname: chroot name

            :param username: [optional] use alternative username

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.ProjectChrootHandle`
                - text fields: "buildroot_pkgs"

        """
        if pkgs is None:
            pkgs = []

        if not username:
            username = self.username

        url = "{0}/coprs/{1}/{2}/modify/{3}/".format(
            self.api_url, username, projectname, chrootname
        )
        data = {
            "buildroot_pkgs": " ".join(pkgs)
        }
        data = self._fetch(url, data=data, method="post")

        response = CoprResponse(
            client=self,
            method="modify_project_chroot_details",
            data=data,
            parsers=[
                fabric_simple_fields_parser(
                    ["buildroot_pkgs", "output", "error"],
                    "BuildDetailsParser"
                )
            ]
        )
        response.handle = ProjectChrootHandle(
            client=self, chrootname=chrootname, username=username,
            projectname=projectname, response=response
        )
        return response

    def search_projects(self, query):
        """ Search projects by substring

            :param query: substring to search

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **projects_list**: list of
                  :py:class:`~.responses.ProjectWrapper`
        """
        url = "{0}/coprs/search/{1}/".format(
            self.api_url, query
        )
        data = self._fetch(url, skip_auth=True)

        response = CoprResponse(
            client=self,
            method="search_projects",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
                ProjectListParser
            ]
        )
        response.handle = BaseHandle(client=self, response=response)
        return response

    def get_build_config(self, project, chroot):
        """
        Return build configuration for given project/chroot.
        :param project: project name, e.g. USER/PROJ, or @GROUP/PROJ.
        :param chroot: chroot name, e.g. fedora-rawhide-x86_64
        :return: :py:class:`~.responses.CoprResponse`
            with additional fields:
            - **build_config**: generated build config contents (dict)
        """
        url = "{0}/coprs/{1}/build-config/{2}".format(
            self.api_url,
            project,
            chroot,
        )
        data = self._fetch(url, skip_auth=True)
        response = CoprResponse(
            client=self,
            method="get_build_config",
            data=data,
            parsers=[
                CommonMsgErrorOutParser,
                BuildConfigParser,
            ]
        )
        return response

    def get_module_repo(self, owner, copr, name, stream, version, arch):
        """ Gets URL to module DNF repository

            :param owner: str owner name (can be user or @group)
            :param copr: str copr name
            :param name: str module name
            :param stream: str module stream
            :param version: int module version
            :param arch: str build architecture

            :return: :py:class:`~.responses.CoprResponse`
                with additional fields:

                - **handle:** :py:class:`~.responses.BaseHandle`
                - text fields: "repo"

        """
        url = "{}/module/repo/".format(self.api_url)
        data = {"owner": owner, "copr": copr, "name": name, "stream": stream, "version": version, "arch": arch}

        fetch = self._fetch(url, data=data, skip_auth=True, method="post")
        response = CoprResponse(
            client=self,
            method="get_module_repo",
            data=fetch,
            parsers=[
                CommonMsgErrorOutParser,
                ProjectListParser
            ]
        )
        response.handle = BaseHandle(client=self, response=response)
        return response

    def build_module(self, modulemd, ownername=None, projectname=None):
        if not ownername:
            ownername = self.username

        url = "{0}/coprs/{1}/{2}/module/build/".format(
            self.api_url, ownername, projectname
        )

        if isinstance(modulemd, io.BufferedIOBase):
            data = {"modulemd": (os.path.basename(modulemd.name), modulemd, "application/x-rpm")}
        else:
            data = {"scmurl": modulemd, "branch": "master"}

        def fetch(url, data, method):
            m = MultipartEncoder(data)
            monit = MultipartEncoderMonitor(m, lambda x: x)
            return self._fetch(url, monit, method="post", headers={'Content-Type': monit.content_type})

        # @TODO Refactor process_package_action to be general purpose
        response = self.process_package_action(url, None, None, data=data, fetch_functor=fetch)
        return response



# coding: utf-8
import six
import json

if six.PY3:
    from unittest import mock
    from unittest.mock import MagicMock
else:
    import mock
    from mock import MagicMock

import pytest

from copr.client_v2.net_client import NetClient, RequestError


class TestNetClient(object):

    def setup_method(self, method):
        # print(sys.path)

        self.base_response = MagicMock()
        self.base_response.status_code = 200
        self.content = {
            "msg": "Lorem ipsum dolor sit amet"
        }
        self.base_response.content = json.dumps(self.content)
        self.base_response.headers = {
            "content-type": "application/json"
        }

        self.login = u"foo"
        self.password = u"bar"
        self.base_url = "http://example.com/abcd"

        self.nc = NetClient()
        self.nc_with_auth = NetClient(self.login, self.password)

    def test_get_simple(self):
        with mock.patch('copr.client_v2.net_client.request') as mc_request:
            mc_request.return_value = self.base_response
            res = self.nc.request(self.base_url)
            # import ipdb; ipdb.set_trace()
            assert res.status_code == self.base_response.status_code
            assert res.json == self.content
            assert res.headers == self.base_response.headers

    def test_unsupported_method(self):
        with mock.patch('copr.client_v2.net_client.request') as mc_request:
            with pytest.raises(RequestError) as exc_info:
                self.nc.request(self.base_url, method="non_existing_method")

            # some coverage for Request error
            assert str(exc_info.value) is not None
            with pytest.raises(ValueError):
                # import ipdb; ipdb.set_trace()
                assert exc_info.value.response_json



from __future__ import absolute_import

from .helpers import config_from_file
from .client import Client
from .proxies import BaseProxy
from .proxies.project import ProjectProxy
from .proxies.build import BuildProxy
from .proxies.package import PackageProxy
from .proxies.mock_chroot import MockChrootProxy
from .proxies.project_chroot import ProjectChrootProxy
from .proxies.build_chroot import BuildChrootProxy
from .proxies.module import ModuleProxy
from .exceptions import (CoprException,
                         CoprRequestException,
                         CoprNoResultException,
                         CoprValidationException,
                         CoprNoConfigException,
                         CoprConfigException)


__all__ = [
    config_from_file,
    Client,

    BaseProxy,
    BuildProxy,
    PackageProxy,
    MockChrootProxy,
    ProjectChrootProxy,
    BuildChrootProxy,
    ModuleProxy,

    CoprException,
    CoprRequestException,
    CoprNoResultException,
    CoprValidationException,
    CoprNoConfigException,
    CoprConfigException,
]



from __future__ import absolute_import

import os
import json
import requests
from munch import Munch
from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor
from .helpers import List
from .exceptions import CoprRequestException, CoprNoResultException, CoprTimeoutException


GET = "GET"
POST = "POST"
PUT = "PUT"


class Request(object):
    # This should be a replacement of the _fetch method from APIv1
    # We can have Request, FileRequest, AuthRequest/UnAuthRequest, ...

    def __init__(self, endpoint, api_base_url=None, method=None, data=None, params=None, auth=None):
        """
        :param endpoint:
        :param api_base_url:
        :param method:
        :param data: dict
        :param params: dict for constructing query params in URL (e.g. ?key1=val1)
        :param auth: tuple (login, token)

        @TODO maybe don't have both params and data, but rather only one variable
        @TODO and send it as data on POST and as params on GET
        """
        self.endpoint = endpoint
        self.api_base_url = api_base_url
        self._method = method or GET
        self.data = data
        self.params = params or {}
        self.auth = auth
        self.headers = None

    @property
    def endpoint_url(self):
        endpoint = self.endpoint.strip("/").format(**self.params)
        return os.path.join(self.api_base_url, endpoint)

    @property
    def method(self):
        return self._method.upper()

    def send(self):
        try:
            response = requests.request(**self._request_params)
        except requests.exceptions.ConnectionError:
            raise CoprRequestException("Unable to connect to {0}.".format(self.api_base_url))
        handle_errors(response)
        return response

    @property
    def _request_params(self):
        return {
            "url": self.endpoint_url,
            "auth": self.auth,
            "json": self.data,
            "method": self.method,
            "params": self.params,
            "headers": self.headers,
        }


class FileRequest(Request):
    def __init__(self, endpoint, files=None, progress_callback=None, **kwargs):
        super(FileRequest, self).__init__(endpoint, **kwargs)
        self.files = files
        self.progress_callback = progress_callback

    @property
    def _request_params(self):
        params = super(FileRequest, self)._request_params

        data = self.files or {}
        data["json"] = ("json", json.dumps(self.data), "application/json")

        callback = self.progress_callback or (lambda x: x)
        m = MultipartEncoder(data)
        params["json"] = None
        params["data"] = MultipartEncoderMonitor(m, callback)
        params["headers"] = {'Content-Type': params["data"].content_type}
        return params


def munchify(response):
    data = response.json()
    if "items" in data:
        return List(items=[Munch(obj) for obj in data["items"]],
                    meta=Munch(data["meta"]), response=response)
    return Munch(data, __response__=response)


def handle_errors(response):
    try:
        response_json = response.json()
        if "error" not in response_json:
            return

        if response.status_code == 404:
            raise CoprNoResultException(response_json["error"], response=response)

        if response.status_code == 504:
            raise CoprTimeoutException(response_json["error"], response=response)

        raise CoprRequestException(response_json["error"], response=response)

    except ValueError:
        raise CoprRequestException("Request is not in JSON format, there is probably a bug in the API code.",
                                   response=response)
